import setuptools
import sys

assert sys.version_info >= (3, 6)

setuptools.setup(
  name='op-dev',
  version='1.0.2',
  author='Alex Exarhos',
  author_email='AlexExarhos@gmail.com',
  packages=[ 'app' ],
  entry_points={
    'console_scripts': [
      'op-dev = app.__main__:main'
    ]
  },
  install_requires=[
    'boto3 >= 1.18.1',
    'op-service-client >= 0.1.0',
    'python-dateutil >= 2.8.1',
    'pytz >= 2021.3',
    'docker >= 4.1.0',
    'gitpython >= 3.0.5',
    'tzlocal >= 2.0.0'
  ]
)
